<?php
declare(strict_types=1);

/* webEdition CMS
 *
 * $Rev: 2207bf7dd286 $
 * $Author: Marc Kraemer <marc.kraemer@aldente-it.de> $
 * $Date: 2020-06-02 13:58 +0200 $
 *
 * This source is part of webEdition CMS. webEdition CMS is
 * free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * any later version.
 *
 * The GNU General Public License can be found at
 * http://www.gnu.org/copyleft/gpl.html.
 * A copy is found in the textfile
 * webEdition/licenses/webEditionCMS/License.txt
 *
 * @license	http://www.gnu.org/copyleft/gpl.html  GPL
 */

/**
 * This class contains all functions needed for the update process
 * TBD if we divide this class in several classes
 */
class liveUpdateFunctions{
	/** @var array<string,string[]> */
	public $QueryLog = [
		'success' => [],
		'tableChanged' => [],
		'error' => [],
		'entryExists' => [],
		'tableExists' => [], //needed from server functions
	];

	/** @var string */
	protected $defaultEngine = '';

	/** @var we_database_base */
	protected $db;
	protected $allowedEngines = ['myisam', 'aria'];

	public function __construct(){
		$this->db = new DB_WE();
		$this->defaultEngine = (string) f('show variables LIKE "default_storage_engine"', 'Value', $this->db);
		if(!in_array(strtolower($this->defaultEngine), $this->allowedEngines)){
			$engines = $this->db->getAllFirstq('SHOW ENGINES');
			foreach(array_keys($engines) as $eng){
				if(strtolower((string) $eng) === 'aria'){
					$this->defaultEngine = 'aria';
					return;
				}
			}
			$this->defaultEngine = 'MyISAM';
		}
	}

	/*
	 * Functions for updatelog
	 */

	function insertUpdateLogEntry(string $action, string $version, int $errorCode): void{
		DB_WE::$inst->query('INSERT INTO ' . UPDATE_LOG_TABLE . ' SET ' . we_database_base::arraySetter([
				'aktion' => $action,
				'versionsnummer' => $version,
				'error' => $errorCode
		]));
	}

	/**
	 * @param string $type
	 * @param string $premessage
	 * @param integer $errorCode
	 * @param string $version
	 */
	function insertQueryLogEntries(string $type, string $premessage, int $errorCode, string $version): void{
		// insert notices first
		if(isset($this->QueryLog[$type])){
			$message = $premessage;

			foreach($this->QueryLog[$type] as $noticeMessage){
				$message .= "<br/>$noticeMessage\n";
			}
			$this->insertUpdateLogEntry($message, $version, $errorCode);
		}
	}

	/**
	 * Decode encoded strings submit from liveupdater
	 *
	 * @param string $string
	 * @return string
	 */
	function decodeCode(string $string): string{
		$string = base64_decode($string);
		if($string && $string[0] === 'x'){
			$str = gzuncompress($string);
			return ($str === false ? $string : $str);
		}
		return $string ?: '';
	}

	/**
	 * prepares given php-code
	 * - replaces doc_root
	 * - edits extension of all containing files
	 *
	 * @return string
	 */
	function preparePhpCode(string $content, string $x = '', string $y = ''): string{
		return $content;
	}

	/**
	 * replaces extension of content
	 *
	 * @param string $content
	 * @param string $needle
	 * @param string $replace
	 * @return string
	 */
	function replaceExtensionInContent(string $content, string $needle, string $replace): string{//FIXME:remove
		return $content;
	}

	function replaceDocRootNeeded(): bool{//FIXME:remove
		return false;
	}

	/**
	 * checks if document root exists, and replaces $_SERVER['DOCMENT_ROOT'] in
	 * $content if needed
	 *
	 * @param string $content
	 * @return string
	 */
	function checkReplaceDocRoot(string $content): string{//FIXME:remove
		//replaces any count of escaped docroot-strings
		return $content;
	}

	/**
	 * fills given array with all files in given dir
	 *
	 * @return string[]
	 */
	function getFilesDir(string $baseDir): array{
		if(!file_exists($baseDir) || !($dh = opendir($baseDir))){
			return [];
		}
		$allFiles = [];
		while(($entry = readdir($dh))){
			if($entry != '' && $entry != '.' && $entry != '..'){
				$entry = $baseDir . '/' . $entry;
				if(!is_dir($entry)){
					$allFiles[] = $entry;
				}

				if(is_dir($entry)){
					$allFiles = array_merge($allFiles, $this->getFilesDir($entry));
				}
			}
		}
		closedir($dh);
		return $allFiles;
	}

	/**
	 * deletes $dir and all files/dirs inside
	 *
	 * @param string $dir
	 */
	function deleteDir(string $dir): bool{
		if(strpos($dir, './') !== false){
			return true;
		}

		if(!file_exists($dir)){
			return true;
		}

		$dh = opendir($dir);
		if(!$dh){
			return true;
		}
			while(($entry = readdir($dh))){
				if($entry != '' && $entry != "." && $entry != '..'){
					$entry = $dir . '/' . $entry;
					if(is_dir($entry)){
						$this->deleteDir($entry);
					} else {
						$this->deleteFile($entry);
					}
				}
			}
			closedir($dh);
			return rmdir($dir);
		}

	/**
	 * Reads filecontent in a string and returns it
	 *
	 * @param string $filePath
	 * @return string
	 */
	function getFileContent(string $filePath): string{
		return (file_exists($filePath) ? file_get_contents($filePath) ?: '' : '');
	}

	/**
	 * writes filecontent in a file
	 *
	 * @param string $filePath
	 * @param string $newContent
	 * @return boolean
	 */
	function filePutContent(string $filePath, string $newContent): bool{
		if($this->checkMakeDir(dirname($filePath)) && $this->checkMakeFileWritable($filePath)){
			$fh = fopen($filePath, 'wb');
			if($fh){
				fwrite($fh, $newContent, strlen($newContent));
				fclose($fh);
				//if we write a php file, invalidate cache if used.
				if(substr($filePath, -4) === '.php' && function_exists('opcache_invalidate')){
					opcache_invalidate($filePath, true);
				}
				//FIXME: current user must write because of "prepare"
				if(!chmod($filePath, 0644)){
					return false;
				}
				return true;
			}
		}
		return false;
	}

	/**
	 * This function checks if a given file is writable and tries to chmod() if necessary
	 *
	 * @param string $filePath
	 * @return boolean
	 */
	function checkMakeFileWritable(string $filePath = '', int $mod = 0750): bool{
		$filePath = LIVEUPDATE_SOFTWARE_DIR . $filePath;

		if(file_exists($filePath) && !is_writable($filePath)){
			if(!chmod($filePath, $mod)){
				return false;
			}
		}

		return true;
	}

	/**
	 * This function checks if given dir exists, if not tries to create it
	 *
	 * @param string $dirPath
	 * @return boolean
	 */
	function checkMakeDir(string $dirPath, int $mod = 0755): bool{
		umask(0022);
		// open_base_dir - seperate document-root from rest
		$dirPath = rtrim(str_replace(['///', '//'], '/', $dirPath), '/');

		$dir = (defined('LIVEUPDATE_SOFTWARE_DIR') ? LIVEUPDATE_SOFTWARE_DIR : WEBEDITION_PATH);
		if(strpos($dirPath, $dir) === 0){
			$preDir = $dir;
			$dir = substr($dirPath, strlen($dir)) ?: '';
		} else {
			$preDir = '';
			$dir = $dirPath;
		}

		$pathArray = explode('/', $dir);
		$path = $preDir;

		foreach($pathArray as $subPath){
			$path .= $subPath;
			if($subPath && !is_dir($path)){
				if(!(file_exists($path) || mkdir($path, $mod))){
					return false;
				}
			}
			$path .= '/';
		}

		if(!is_writable($dirPath)){
			if(!chmod($dirPath, $mod)){
				return false;
			}
		}
		return true;
	}

	/**
	 * @param string $file
	 * @return boolean true if the file is not existent after this call
	 */
	function deleteFile(string $file): bool{
		return (file_exists($file) ? @unlink($file) : true);
	}

	/**
	 * moves $source file to new $destination
	 *
	 * @param string $source
	 * @param string $destination
	 * @return boolean false if move was not successful
	 */
	function moveFile(string $source, string $destination): bool{

		if($source == $destination || !file_exists($source)/* happens if update is retriggered */){
			return true;
		}
		if($this->checkMakeDir(dirname($destination))){

			if(filesize($source) == 0){//assume error, add warning, keep file!
				$this->QueryLog['error'][] = 'File ' . $source . ' was empty, not overwriting!';
				//keep going
				return true;
			}

			if(!isset($_SESSION['weS']['moveOk'])){
				touch($source . 'x');
				$_SESSION['weS']['moveOk'] = @rename($source . 'x', $destination . 'x');
				$this->deleteFile($destination . 'x');
				$this->deleteFile($source . 'x');
				$this->insertUpdateLogEntry('Using ' . ($_SESSION['weS']['moveOk'] ? 'move' : 'copy') . ' for installation', WE_VERSION, 0);
			}
			if(substr($destination, -4) === '.php' && function_exists('opcache_invalidate')){
				opcache_invalidate($destination, true);
			}

			if($_SESSION['weS']['moveOk']){
				return rename($source, $destination);
			}
			//rename seems to have problems - we do it old school way: copy, on success delete
			if($this->deleteFile($destination)){
				if(copy($source, $destination)){
					$this->deleteFile($source);
					//should we handle file deletion?
					return true;
				}
			}
		}
		return false;
	}

	/**
	 * returns if selected file is a php file or not, important also check html files
	 *
	 * @param string $path
	 * @return boolean
	 */
	function isPhpFile(string $path): bool{

		$matches = [];
		if(preg_match('/\.([^\..]+)$/', $path, $matches)){
			switch(strtolower($matches[1])){
				case 'jpg':
				case 'gif':
				case 'jpeg':
				case 'sql':
					return false;
			}
		}
		return true;
	}

	/**
	 * This file searchs $needle in given file and replaces it with $replace
	 * If needle is empty the whole file is overwritten. Also
	 * $_SERVER[DOCUMENT_ROOT] is replaced if necessary
	 *
	 * @param string $filePath
	 * @param string $replace
	 * @param string $needle
	 * @return boolean
	 */
	function replaceCode(string $filePath, string $replace, string $needle = ''): bool{
		if(strpos($filePath, 'we/include/we_version') !== false || !$this->replaceDocRootNeeded()){
			return true;
		}

		if(file_exists($filePath)){
			$needle = $this->decodeCode($needle);
			$replace = $this->decodeCode($replace);
			$oldContent = $this->getFileContent($filePath);
			$replace = $this->checkReplaceDocRoot($replace);
			$newContent = ($needle ? preg_replace('/' . preg_quote($needle) . '/', $replace, $oldContent) : $replace );

			return ($this->filePutContent($filePath, $newContent));
		}
		return false;
	}

	/*
	 * Functions for patches
	 */

	/**
	 * executes patch.
	 *
	 * @param string $path
	 * @return boolean
	 */
	function executePatch(string $path): bool{
		include_once($path);
		return true;
	}

	/**
	 * returns array with all columns of given tablename
	 *
	 * @param string $tableName
	 * @return array<string,array{Field:string,Type:string,Null:string,Key:string,Default:string,Extra:string,Pos:int}>
	 */
	function getFieldsOfTable(string $tableName, bool $lower = false): array{
		$db = $this->db;

		$fieldsOfTable = [];
		$db->query('DESCRIBE ' . $db->escape($tableName));
		$pos = 0;
		while($db->next_record()){
			$fieldsOfTable[$lower ? strtolower($db->f('Field')) : $db->f('Field')] = [
				'Field' => $db->f('Field'),
				'Type' => $db->f('Type'),
				'Null' => $db->f('Null'),
				'Key' => $db->f('Key'),
				'Default' => $db->f('Default'),
				'Extra' => $db->f('Extra'),
				'Pos' => $pos++,
			];
		}
		return $fieldsOfTable;
	}

	/**
	 * returns array with key information of a table by tablename
	 *
	 * @param string $tableName
	 * @return array<string,array<string,string>>
	 */
	function getKeysFromTable(string $tableName, bool $lowerKeys = false): array{
		$db = $this->db;
		$keysOfTable = [];
		$keys = $db->getAllq('SHOW INDEX FROM ' . $db->escape($tableName));
		foreach($keys as $key){
			if($key['Key_name'] === 'PRIMARY'){
				$indexType = 'PRIMARY';
			} else if(in_array('FULLTEXT', [$key['Comment'], $key['Index_type']], true)){// this also depends on mysqlVersion
				$indexType = 'FULLTEXT';
			} else if($key['Non_unique'] == 0){
				$indexType = 'UNIQUE';
			} else {
				$indexType = 'INDEX';
			}

			$keyname = $lowerKeys ? strtolower($key['Key_name']) : $key['Key_name'];

			if(!isset($keysOfTable[$keyname]) || !in_array($indexType, $keysOfTable[$keyname], true)){
				$keysOfTable[$keyname]['index'] = $indexType;
			}
			$keysOfTable[$keyname][$key['Seq_in_index']] = $key['Column_name'] . ($key['Sub_part'] ? '(' . $key['Sub_part'] . ')' : '');
		}
		return $keysOfTable;
	}

	/**
	 * expects array from getFieldsOfTable and returns generated queries to
	 * alter these fields
	 *
	 * @param array<string,array{array{Field:string,Type:string,Null:string,Default:string,Extra:string,Pos:int,last:string},bool}> $fields
	 * @param string $tableName
	 * @return string[]
	 */
	function getAlterTableForFields(array $fields, string $tableName, string $origQuery): array{
		$queries = [];
		$queryArray = array_map('trim', explode("\n", $origQuery));
		foreach($fields as $fieldName => $fieldstat){
			list($fieldInfo, $isNew) = $fieldstat;
			$pos = ($fieldInfo['last'] === 'FIRST' ? $fieldInfo['last'] : 'AFTER `' . $fieldInfo['last'] . '`');
			$extra = strtoupper($fieldInfo['Extra']) . ' ' . $pos;
			//note: auto_increment cols must have an index!
			if(strpos($extra, 'AUTO_INCREMENT') !== false){
				$keyfound = false;
				$Currentkeys = $this->getKeysFromTable($tableName);
				foreach($Currentkeys as $ckeys){
					foreach($ckeys as $k){
						if(stripos($k, $fieldName) !== false){
							$keyfound = true;
						}
					}
				}
				if(!$keyfound){
					$extra .= ', ADD INDEX _temp (`' . $fieldInfo['Field'] . '`)';
				}
			}

			//if(in_array(strtoupper($fieldInfo['Extra']), ['PERSISTENT', 'STORED', 'VIRTUAL', 'VIRTUAL GENERATED', 'VIRTUAL STORED'])){
			//Virtual columns can not be extracted from show table :(
			foreach($queryArray as $q){
				if(preg_match('|^`?' . $fieldInfo['Field'] . '`? |', $q)){
					$queries[] = ($isNew ? 'ADD ' : 'MODIFY ') . trim($q, ',') . ' ' . $pos;
				}
			}
		}
		return $queries;
	}

	/**
	 * returns array with queries to update keys of table
	 *
	 * @param array<string,array{string[],bool}> $fields
	 * @return array{string[],string[]}
	 */
	function getAlterTableForKeys(array $fields): array{
		$drop = $add = [];

		foreach($fields as $key => $indexestat){
			list($indexes, $isNew) = $indexestat;
			//escape all index fields
			$indexes = array_map('addslashes', $indexes);

			$type = $indexes['index'];
			$mysl = '`';
			if($type === 'PRIMARY'){
				$key = 'KEY';
				$mysl = '';
			}
			//index is not needed any more and disturbs implode
			unset($indexes['index']);
			$myindexes = [];
			foreach($indexes as $index){
				if(strpos($index, '(') === false){
					$myindexes[] = '`' . $index . '`';
				} else {
					$myindexes[] = '`' . str_replace('(', '`(', $index);
				}
			}
			if(!$isNew){
				$drop[] = 'DROP ' . ($type === 'PRIMARY' ? $type : 'INDEX') . ' ' . $mysl . $key . $mysl;
			}
			$add[] = 'ADD ' . $type . ' ' . $mysl . $key . $mysl . ' (' . implode(',', $myindexes) . ')';
		}
		return [$drop, $add];
	}

	function getTableEngine(string $table): string{
		return f('SHOW TABLE STATUS LIKE "' . $table . '"', 'Engine') ?: 'NOT_FOUND';
	}

	function getRowFormat(string $table): string{
		return f('SHOW TABLE STATUS LIKE "' . $table . '"', 'Row_format') ?: 'NOT_FOUND';
	}

	/**
	 * executes all queries in a single file
	 * - there is one query, if create-statement
	 * - many queris, if insert statements
	 *
	 *
	 * @param string $path
	 * @return boolean
	 */
	function executeQueriesInFiles(string $path): bool{
		$data = (string) str_ireplace('ENGINE=MyISAM', 'ENGINE=' . $this->defaultEngine, $this->getFileContent($path));
		$queries = explode('/* query separator */', $data);
		foreach($queries as $query){
			if(!$this->executeUpdateQuery($query)){//don't continue current file in case of errors
				return false;
			}
		}

		return true;
	}

	/**
	 * updates the database with given dump.
	 *
	 * @param string $query
	 */
	function executeUpdateQuery(string $query): bool{
		$db = $this->db;

		// when executing a create statement, try to create table,
		// change fields when needed.

		if(strpos($query, '###INSTALLONLY###') !== false){// potenzielles Sicherheitsproblem, nur im LiveUpdate nicht ausführen
			return true;
		}

		$query = strtr(trim($query), ['###TBLPREFIX###' => defined('LIVEUPDATE_TABLE_PREFIX') && LIVEUPDATE_TABLE_PREFIX ? LIVEUPDATE_TABLE_PREFIX : TBL_PREFIX, '###UPDATEONLY###' => '']);
		$matches = [];
		if(preg_match('/###UPDATEDROPCOL\(([^,]*),([^)]*)\)###/', $query, $matches)){
			$query = ($db->isColExist($matches[2], $matches[1]) ? 'ALTER TABLE ' . $db->escape($matches[2]) . ' DROP COLUMN `' . $db->escape($matches[1]) . '`' : '');
		}
		if(preg_match('/###ONCOL\(([^,]*),([^)]*)\)(.+);###/', $query, $matches)){
			$query = ($db->isColExist($matches[2], $matches[1]) ? $matches[3] : '');
		}
		if(preg_match('/###ONNOTCOL\(([^)]*)\)(.+);###/', $query, $matches)){
			$cols = explode(',', $matches[1]);
			$table = (string) array_pop($cols);
			if(!$db->isTabExist($table)){
				return true;
			}
			$ok = true;
			foreach($cols as $col){
				$ok = $ok && (!$db->isColExist($table, $col));
			}
			$query = ($ok ? $matches[2] : '');
		}
		//handle if key is not set, should be used after table def. so handling code, e.g. truncate, copy... can be put here
		if(preg_match('/###ONKEYFAILED\(([^,]+),([^)]+)\)([^#]+)###/', $query, $matches)){
			$keys = $this->getKeysFromTable($matches[2]);
			$query = (!isset($keys[$matches[1]]) ? $matches[3] : '');
		}
		if(preg_match('/###UPDATEDROPKEY\(([^,]+),([^)]+)\)###/', $query, $matches)){
			$db->query('SHOW KEYS FROM ' . $db->escape($matches[2]) . ' WHERE Key_name="' . $matches[1] . '"');
			$query = ($db->num_rows() ? 'ALTER TABLE ' . $db->escape($matches[2]) . ' DROP KEY ' . $db->escape($matches[1]) : '');
		}
		if(preg_match('/###ONTAB\(([^)]*)\)(.+);###/', $query, $matches)){
			$query = ($db->isTabExist($matches[1]) ? $matches[2] : '');
		}

		if(stripos($query, 'CREATE TABLE ') === 0){
			$query = preg_replace('/;$/', we_database_base::getCharsetCollation() . ';', $query, 1);
		}

		$tabExists = false;
		if(preg_match('/CREATE TABLE (\w+) \(/', $query, $matches)){
			if($db->isTabExist($matches[1])){//tab exists
				$db->Errno = 1050;
				$tabExists = true;
			}
		}

		if(!$query || (!$tabExists && $db->query($query))){
			return true;
		}

		switch($db->Errno){
			case 1050: // this table already exists
				// the table already exists,
				// make tmptable and check these tables
				$namePattern = '/CREATE TABLE (\w+) \(/';
				preg_match($namePattern, $query, $matches);
				if($matches[1]){

					// get name of table and build name of temptable
					// realname of the new table
					$tableName = $matches[1];

					// tmpname - this table is to compare the incoming dump
					// with existing table
					$tmpName = '__we_delete__' . trim($tableName, '`');
					$backupName = trim($tableName, '`') . '_backup';

					$db->query('DROP TABLE IF EXISTS ' . $db->escape($tmpName)); // delete table if already exists
					$db->query('DROP TEMPORARY TABLE IF EXISTS ' . $db->escape($tmpName)); // delete table if already exists
					$db->query('DROP TABLE IF EXISTS ' . $db->escape($backupName)); // delete table if already exists
					//update enigine of table if it does not match
					if(!in_array(strtolower($this->getTableEngine($tableName)), $this->allowedEngines, true)){
						$db->query('ALTER TABLE `' . $tableName . '` ENGINE=' . $this->defaultEngine);
					}
					$currentRowFormat = $this->getRowFormat($tableName);
					$db->query('SHOW CREATE TABLE ' . $db->escape($tableName));
					list(, $orgTable) = ($db->next_record(MYSQLI_NUM) ? $db->getRecord() : ['', '']);
					$orgTable = preg_replace($namePattern, 'CREATE TABLE ' . $db->escape($backupName) . ' (', $orgTable);
					$origTable = $this->getFieldsOfTable($tableName, true);

					// create temptable
					$db->query('DROP TEMPORARY TABLE IF EXISTS ' . $db->escape($tmpName));
					$tmpQuery = preg_replace($namePattern, 'CREATE TEMPORARY TABLE ' . $db->escape($tmpName) . ' (', $query);
					$db->query(trim($tmpQuery));
					$newRowFormat = $this->getRowFormat($tableName);
					if($currentRowFormat != $newRowFormat){
						$db->query('ALTER TABLE `' . $tableName . '` ROW_FORMAT=' . $newRowFormat);
					}
					$newTable = $this->getFieldsOfTable($tmpName);

					// get information from existing and new table
					if(empty($newTable)){
						$this->QueryLog['error'][] = 'Update of table ' . $tableName . " failed (create temporary table was not successfull)\n-- $tmpQuery --";
						break;
					}

					// get keys from existing and new table
					$origTableKeys = $this->getKeysFromTable($tableName, true);
					$newTableKeys = $this->getKeysFromTable($tmpName);
					// determine changed and new fields.
					$changeFields = []; // array with new/changed fields
					$lastField = 'FIRST';
					foreach($newTable as $fieldName => $newField){
						$newField['last'] = $lastField;
						$lowField = strtolower((string) $fieldName);
						if(isset($origTable[$lowField])){ // field exists
							$orgField = $origTable[$lowField];
							if(!(
								$newField['Field'] === $orgField['Field'] && //e.g. case changed
								$newField['Type'] === $orgField['Type'] &&
								$newField['Null'] === $orgField['Null'] &&
								$newField['Default'] === $orgField['Default'] &&
								$newField['Extra'] === $orgField['Extra'] &&
								$newField['Pos'] === $orgField['Pos'])){
								$changeFields[$fieldName] = [$newField, false];
							}
						} else { // field does not exist
							$changeFields[$fieldName] = [$newField, true];
						}
						$lastField = $fieldName;
					}

					// determine new keys
					// get all queries to add/change fields, keys
					$changeFieldsQuery = ($changeFields ? $this->getAlterTableForFields($changeFields, $tableName, $query) : []);

					$changedKeys = [];
					foreach($newTableKeys as $keyName => $indexes){
						$lkeyName = strtolower($keyName);
						if(isset($origTableKeys[$lkeyName])){
							//index-type changed
							if($origTableKeys[$lkeyName]['index'] != $indexes['index']){
								$changedKeys[$keyName] = [$indexes, false];
								continue;
							}

							for($i = 1; $i < count($indexes); $i++){
								if(!in_array($indexes[$i], $origTableKeys[$lkeyName], true)){
									$changedKeys[$keyName] = [$indexes, false];
									break;
								}
							}
						} else {
							$changedKeys[$keyName] = [$indexes, true];
						}
					}
					// get all queries to change existing keys
					if(empty($changedKeys)){
						$allAlterQ = $changeFieldsQuery;
						$drop = [];
					} else {
						list($drop, $add) = $this->getAlterTableForKeys($changedKeys);
						$allAlterQ = array_merge($drop, $changeFieldsQuery, $add);
					}

					//clean-up, if there is still a temporary index - make sure this is the first statement, since new temp might be created
					if(isset($origTableKeys['_temp'])){
						$allAlterQ = array_merge(['DROP INDEX _temp'], $allAlterQ);
					}
					if($allAlterQ){
						// execute all queries
						$success = true;
						$duplicate = false;
						//disable strict mode!!! The problem exists, if we convert e.g. text-columns to int, and the value is "", mysql refuses to change the column cause this will destroy data.
						$db->query('SET SESSION sql_mode=""');
						//try full update
						$query = 'ALTER TABLE `' . $tableName . '` ' . implode(',', $allAlterQ);
						if($db->query($query)){
							$this->QueryLog['success'][] = $query;
						} else {
							//unknown why mysql don't show correct error
							switch($db->Errno){
								case 1062:
								case 0:
									$duplicate = true;
									$this->QueryLog['tableChanged'][] = $tableName;
									break;
								default:
									$this->QueryLog['error'][] = $db->Errno . ' ' . urlencode($db->Error) . "\n-- $query --";
							}
							$success = false;
						}

						if($success){
							$this->QueryLog['tableChanged'][] = $tableName . "\n<!-- $query -->";
						} else if($duplicate){
							//no alter has yet been run, the table is not changed!
							//run only updates without keys!
							//we force this change, it only fails if charset is changed on single columns
							//if we have dropped keys we need to drop them before
							$query = 'ALTER TABLE `' . $tableName . '` ' . implode(',', array_merge($drop ?? [], $changeFieldsQuery));
							if(!$db->query($query)){
								$this->QueryLog['error'][] = $query;
								return false;
							}
							//now move to backup
							if($db->query('RENAME TABLE ' . $db->escape($tableName) . ' TO ' . $db->escape($backupName))){
								//create an empty table which looks like the old one
								$db->query($orgTable);
								$db->lock([$tableName => 'write', $backupName => 'read']);
								//apply everything to the new table what failed on the first one. Note this table is empty, so index should not fail now.
								$query = 'ALTER TABLE `' . $tableName . '` ' . implode(',', $allAlterQ);
								if(!$db->query($query)){
									$this->QueryLog['error'][] = $db->Errno . ' ' . urlencode($db->Error) . "\n-- $query --";
								}
								//add all data from the original table. If we have duplicate keys, we use the first match, not the last!
								$db->query('INSERT IGNORE INTO ' . $db->escape($tableName) . ' SELECT * FROM ' . $db->escape($backupName));
								$db->unlock();
							}
						}
						$SearchTempTableKeys = $this->getKeysFromTable($tableName);
						if(isset($SearchTempTableKeys['_temp'])){
							$db->query('ALTER TABLE ' . $db->escape($tableName) . ' DROP INDEX _temp');
						}
					} else {
						//$this->QueryLog['tableExists'][] = $tableName;
					}

					$db->query('DROP TABLE IF EXISTS ' . $db->escape($tmpName));
				}
				break;
			case 1062:
				$this->QueryLog['entryExists'][] = $db->Errno . ' ' . urlencode($db->Error) . "\n<!-- $query -->";
				return false;
			case 0:
			case 1065:
				//ignore empty queries
				return true;
			default:
				$this->QueryLog['error'][] = $db->Errno . ' ' . urlencode($db->Error) . "\n-- $query --";
				return false;
		}

		return true;
	}

	/**
	 * returns log array for db-queries
	 * @return array
	 */
	function getQueryLog(string $specific = ''): array{
		return ($specific ? $this->QueryLog[$specific] : $this->QueryLog);
	}

	/**
	 * resets query log, this is done after each query file.
	 */
	function clearQueryLog(): void{
		foreach($this->QueryLog as &$cur){
			$cur = [];
		}
	}

	/**
	 * returns array with all installed languages
	 *
	 * @return string[]
	 */
	public static function getInstalledLanguages(): array{
		clearstatcache();

		//	Get all installed Languages
		$language_directory = dir(LIVEUPDATE_LANGUAGE_ROOTDIR);
		if(!$language_directory){
			return [];
		}
		//	Get all installed Languages
		$installedLanguages = [];
		while(false !== ($entry = $language_directory->read())){
			if($entry != '.' && $entry != '..'){
				if(is_dir(LIVEUPDATE_LANGUAGE_ROOTDIR . $entry)){
					$installedLanguages[] = $entry;
				}
			}
		}
		$language_directory->close();

		return $installedLanguages;
	}

	function removeObsoleteFiles(string $path): bool{
		if(is_file($path . 'del.files')){
			$delFiles = [];
			if(($all = file($path . 'del.files', FILE_IGNORE_NEW_LINES))){
				foreach($all as $cur){
					$recursive = false;
					if($cur[0] === '!'){
						$cur = substr($cur, 1);
						$recursive = true;
					}
					if(file_exists(WEBEDITION_PATH . $cur)){
						if(is_file(WEBEDITION_PATH . $cur)){
							$delFiles[] = $cur;
							unlink(WEBEDITION_PATH . $cur);
						} elseif(is_dir(WEBEDITION_PATH . $cur)){
							$delFiles[] = 'Folder: ' . $cur;
							we_base_file::deleteLocalFolder(WEBEDITION_PATH . $cur, $recursive);
						}
					}
				}
			}
			unlink($path . 'del.files');
			file_put_contents($path . 'deleted.files', ($all ? "Deleted Files: " . count($delFiles) . "\n\n" . implode("\n", $delFiles) : "File del.files empty"));
		}

		return true;
	}

	/**
	 *
	 * @param string $what
	 * @param array|array{what:string,pos:int,max:int,text:string} $progress
	 * @return array|array{what:string,pos:int,max:int,text:string}
	 */
	static function weUpdaterDoUpdate(string $what, array $progress = []): array{
		if(method_exists('we_update_updater', 'doUpdate')){
			return we_update_updater::doUpdate($what, $progress);
		}

		return [];
	}

	function removeDirOnlineInstaller(): bool{
		if(is_dir(WE_DOCUMENT_ROOT . '/OnlineInstaller')){
			we_base_file::deleteLocalFolder(WE_DOCUMENT_ROOT . '/OnlineInstaller', true);
		}

		return true;
	}

	/**
	 * This file sets another errorhandler - to make specific error-messages
	 *
	 * @param integer $errno
	 * @param string $errstr
	 * @param string $errfile
	 * @param integer $errline
	 * @param array $errcontext
	 */
	static function liveUpdateErrorHandler(int $errno, string $errstr, string $errfile, int $errline, array $errcontext): bool{

		$GLOBALS['liveUpdateError'] = [
			'errorNr' => $errno,
			'errorString' => $errstr,
			'errorFile' => $errfile,
			'errorLine' => $errline,
		];
		if(method_exists('we_base_errorHandler', 'errorHandler')){
			if(strpos($errstr, 'MYSQL-ERROR') !== 0){
				//don't handle mysql errors, they're handled by updatelog - since some of them are "wanted"
				//log errors to system log, if we have one.
				return we_base_errorHandler::errorHandler($errno, $errstr, $errfile, $errline, $errcontext);
			}
		}
		return true;
	}

}
