<?php
declare(strict_types=1);

/**
 * webEdition CMS
 *
 * $Rev: 3774dbcab28a $
 * $Author: Marc Kraemer <marc.kraemer@aldente-it.de> $
 * $Date: 2020-04-21 15:07 +0200 $
 *
 * This source is part of webEdition CMS. webEdition CMS is
 * free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * any later version.
 *
 * The GNU General Public License can be found at
 * http://www.gnu.org/copyleft/gpl.html.
 * A copy is found in the textfile
 * webEdition/licenses/webEditionCMS/License.txt
 *
 * @license	http://www.gnu.org/copyleft/gpl.html  GPL
 */
abstract class we_backup_cmd implements we_interface_cmd{
	/** @var string */
	private static $js = '';

	public static function we_cmd(string $cmd): bool{
		switch($cmd){
			case 'we_backup_showLog':
				we_backup_wizard::showLog();
				return true;
			case 'we_backup_cmd':
				we_backup_cmd::getHTML();
				return true;
		}
		return false;
	}

	public static function getHTML(): void{
		if(function_exists('apache_setenv')){
			apache_setenv('no-gzip', '1');
		}
		ini_set('zlib.output_compression', '0');

		$cmd = we_base_request::string('cmd');
		if(!$cmd){
			t_e('called without command');
			exit();
		}


		switch($cmd){
			case 'export':
			case 'import':
				try{
					$settings = we_backup_settings::i();

					$last = $settings->limits['requestTime'];
					$settings->limits['requestTime'] = ($_SERVER['REQUEST_TIME'] ??
						//we don't have the time of the request, assume some time is already spent.
						time() - 3);

					if(we_base_request::bool('reload')){
						$tmp = $settings->limits['requestTime'] - $last;
						t_e('Backup caused reload', $last, $settings->limits, $tmp);
						$tmp -= 4;
						if($tmp > 0 && $tmp < $settings->limits['exec']){
							$settings->limits['exec'] = $tmp;
						}

						++$settings->retry;

						if($settings->retry > 10 /* || $settings->options[we_backup_settings::OPTION_BACKUP_COMPRESS] != we_backup_util::NO_COMPRESSION */){//in case of compression the file can't be used
							$settings->retry = 1;
							$jsCmd = new we_base_jsCmd();
							$jsCmd->addMsg(g_l('backup', '[error_timeout]'), we_base_util::WE_MESSAGE_ERROR);
							echo we_html_tools::getHtmlTop('', $jsCmd->getCmds(), we_html_element::body());
							return;
						}
					}
					$settings->limits['lastMem'] = 0;

					self::$js = we_html_element::jsScript(JS_DIR . 'backup_wizard.js', '', ['id' => 'loadVarBackup_wizard',
							'data-backup' => setDynamicVar([
								'reload' => (we_base_request::int('reload') + 1),
								'mode' => (we_base_request::string('cmd') == we_backup_wizard::RECOVER ? 'import' : 'export'),
								'reloadTimer' => (min($settings->limits['exec'], 32) * 1000) + 5000 //wait extra 5 secs
					])]);
				} catch (Exception $ex){
					self::$js = we_html_element::jsScript(JS_DIR . 'backup_wizard.js');
				}
		}

		switch($cmd){
			case 'export':
				self::export();
				we_backup_util::writeLog();
				session_write_close();
				break;
			case 'import':
				self::import();
				break;
			case 'rebuild':
				$jsCmd = new we_base_jsCmd();
				$jsCmd->addCmd('rebuild');
				echo we_html_tools::getHtmlTop('', self::$js . $jsCmd->getCmds(), we_html_element::body());
				break;

			default:
		}
	}

	private static function logExportStart(): void{
		we_backup_util::limitsReached('', 1);
		we_backup_util::delLog();
		we_backup_util::addLog('Start backup export');
		we_backup_settings::i()->logSettingsExport();
		we_backup_util::addLog('Backup steps: FAST_BACKUP');
		we_backup_util::writeLog();
	}

	private static function exportExtern(): void{
		$settings = we_backup_settings::i();
		$exportClass = $settings->class;
		'@phan-var we_backup_exportBase $exportClass';
		if(!($exportClass->open())){
			throw new we_exception_file('error opening file');
		}
		$settings->backup_steps = 2;
		do{
			$start = microtime(true);
			for($i = 0; $i < $settings->backup_steps; $i++){

				$file_to_export = array_pop($settings->extern_files);
				if(empty($file_to_export)){
					break 2;
				}
				we_backup_util::addLog('Exporting file ' . $file_to_export);
				we_backup_util::writeLog();
				$exportClass->writeBinFile($file_to_export, we_backup_exportBase::EXTERNALS_DIR);
			}
			we_backup_util::writeLog();
		} while(!empty($settings->extern_files) && we_backup_util::limitsReached('', intval(microtime(true) - $start)));
		$exportClass->close();
	}

	private static function issueNext(string $type, string $description): void{
		we_backup_util::addLog('Issuing next request.', true);
		switch($type){
			case 'export':
				$percent = we_backup_util::getExportPercent();
				break;
			case 'import':
				$percent = we_backup_util::getImportPercent();
				break;
			default:
				$percent = 0;
		}
		echo we_html_tools::getHtmlTop('', self::$js, we_html_element::body(['onload' => "run('" . $type . "'," . $percent . ",'" . $description . "');"]));

		return;
	}

	private static function backupData(): bool{
		$settings = we_backup_settings::i();
		$settings->backup_steps = 10;
		$settings->class->open();

		do{
			$start = microtime(true);
			for($i = 0; $i < $settings->backup_steps; $i++){
				//t_e($settings);
				if(we_backup_export::export() === false){
// force end
					$settings->row_counter = $settings->row_count;
					return false;
				}
			}
			we_backup_util::writeLog();
		} while(we_backup_util::limitsReached($settings->current_table, intval(microtime(true) - $start)));
		$settings->class->close();
		return true;
	}

	private static function export(): void{
		try{
			$settings = we_backup_settings::i();
		} catch (Exception $ex){
			if(we_backup_preparer::prepareExport() === true){
				self::logExportStart();
				self::issueNext('export', g_l('backup', '[working]'));
			} else {
				t_e('No write permissions!');
			}
			return;
		}
		if($settings->extern_files){
			self::exportExtern();
			self::issueNext('export', g_l('backup', '[external_backup]'));
			return;
		}
		if(self::backupData()){
			self::issueNext('export', we_backup_util::getDescription($settings->current_table, 'export'));
			return;
		}


// export settings from the file
		if($settings->isHandled(we_backup_settings::OPTION_BACKUP_SETTINGS)){
			we_backup_util::addLog('Exporting settings');
			$files = we_backup_util::getSettingsFiles(false);
			we_backup_util::exportFiles($files, we_backup_exportBase::CONF_DIR, true);
		}

		if($settings->isHandled(we_backup_settings::OPTION_BACKUP_HOOKS)){
			$files = [];
			foreach(we_backup_preparer::getFileList(WE_CUSTOM_PATH . we_hook_handler::DIR) as $cur){
				$files[basename($cur)] = $cur;
			}
			we_backup_util::exportFiles($files, we_backup_exportBase::HOOK_DIR, true);
		}

		if($settings->isHandled(we_backup_settings::OPTION_BACKUP_CUSTOMTAGS)){
			$files = [];
			foreach(we_backup_preparer::getFileList(WE_CUSTOM_PATH . we_weTag_tag::DIR) as $cur){
				$files['code/' . basename($cur)] = $cur;
			}
			foreach(we_backup_preparer::getFileList(WE_CUSTOM_PATH . we_weTag_twbase::DIR) as $cur){
				$files['wizzard/' . basename($cur)] = $cur;
			}
			we_backup_util::exportFiles($files, we_backup_exportBase::TAG_DIR, true);
		}

		we_backup_util::writeLog();

		$settings->class->finish();

		echo we_html_tools::getHtmlTop('', we_html_element::jsScript(JS_DIR . 'backup_wizard.js'), we_html_element::body(['onload' => "backupFinished('" . g_l('backup', '[finished]') . "');"]));
	}

	private static function logImportStart(): void{
		we_backup_util::delLog();
		we_backup_util::addLog('Start backup import');
		we_backup_settings::i()->logSettingsImport();
		we_backup_util::writeLog();
	}

	private static function setImportVars(): bool{
		if(we_backup_preparer::prepareImport() === true){

			if(!we_base_file::hasZip()){
				$err = we_backup_preparer::getErrorMessage();
				we_backup_settings::clearSessionData();
				echo we_html_tools::getHtmlTop('', $err, we_html_element::body());
				return false;
			}
			self::logImportStart();
			return true;
		}

		$err = we_backup_preparer::getErrorMessage();

		we_backup_util::writeLog();
		we_backup_settings::clearSessionData();
		echo we_html_tools::getHtmlTop('', $err, we_html_element::body());
		return false;
	}

	private static function delFiles(): void{
		$settings = we_backup_settings::i();
		we_backup_util::addLog('Clean old files: (' . ($settings->files_to_delete_count - count($settings->files_to_delete)) . '/' . $settings->files_to_delete_count . ')');
		do{
			$start = microtime(true);
			for($i = 0; $i < 50; ++$i){
				$file = array_pop($settings->files_to_delete);
				if(!$file){
					break;
				}
				we_base_file::delete($file);
			}
		} while(!empty($settings->files_to_delete) && we_backup_util::limitsReached('', intval(microtime(true) - $start)));
		we_backup_util::addLog('Left old files: ' . count($settings->files_to_delete));
	}

	private static function importData(): void{
		$settings = we_backup_settings::i();
		$class = $settings->class;
		'@phan-var we_backup_importBase $class';
		try{
			$class->open();
		} catch (we_exception_file $e){
			we_backup_util::addLog(sprintf('File %s not readable.', $settings->backup_file), true);
			throw new we_exception_file('File %s not readable: ' . $settings->backup_file);
		}
		do{
			$start = microtime(true);
			we_backup_import::import();
			we_backup_util::writeLog();
		} while($class->canReadMore() && we_backup_util::limitsReached('', intval(microtime(true) - $start)));
		$class->close();
	}

	private static function importCallUpdater(): void{
		$settings = we_backup_settings::i();
		if(empty($settings->update)){
			we_backup_util::addLog('Call updater to fix tables (start)', true);
			$settings->update = we_update_updater::doUpdate('internal');
		} else {
			we_backup_util::addLog('Call updater to fix tables (' . $settings->update['what'] . ')', true);
			// perform update
			$settings->update = we_update_updater::doUpdate((string) $settings->update['what'], $settings->update);
		}
		if(!$settings->update){//no more update changes
			//make sure we_update is run on next request
			$settings->offset++;
		}
	}

	private static function importFinish(): void{
		$settings = we_backup_settings::i();
		we_base_file::deleteLocalFile(TEMP_PATH . $settings->backup_file);

// reload user prefs
		$_SESSION['prefs'] = we_users_user::readPrefs(we_users_user::getID());
		$jsCmd = new we_base_jsCmd();
		$jsCmd->addCmd('importFinished', [
			'doRebuild' => $settings->option(we_backup_settings::OPTION_BACKUP_REBUILD),
			'file' => $settings->option(we_backup_settings::OPTION_BACKUP_REBUILD) ? $settings->backup_file : ''
		]);
		$settings->class->finish();

		list($cmd, $val) = we_gui_headermenu::getMenuReloadCode();
		$jsCmd->addCmd($cmd, $val);

		echo we_html_tools::getHtmlTop('', we_html_element::jsScript(JS_DIR . 'backup_wizard.js') . $jsCmd->getCmds(), we_html_element::body());
		we_backup_util::addLog('Backup import finished', true);
	}

	private static function import(): void{
		try{
			$settings = we_backup_settings::i();
		} catch (Exception $ex){
			if(self::setImportVars()){
				self::issueNext('import', g_l('backup', '[working]'));
			}
			return;
		}

		if(!empty($settings->files_to_delete)){
			self::delFiles();
			self::issueNext('import', g_l('backup', '[delete_old_files]'));
			return;
		}
		if(($settings->offset < $settings->offset_end)){
			we_backup_util::addLog('offset ' . $settings->offset . '/' . $settings->offset_end, true);
			self::importData();
			self::issueNext('import', we_backup_util::getDescription($settings->current_table, 'import'));
			return;
		}
		if(($settings->offset == $settings->offset_end)){
			self::importCallUpdater();
			self::issueNext('import', 'Update ' . (!empty($settings->update['text']) ? $settings->update['text'] : ''));
			return;
		}
		self::importFinish();
	}

}
