<?php
declare(strict_types=1);

/**
 * webEdition CMS
 *
 * $Rev: a1df505902cd $
 * $Author: Marc Kraemer <marc.kraemer@aldente-it.de> $
 * $Date: 2020-06-19 19:17 +0200 $
 *
 * This source is part of webEdition CMS. webEdition CMS is
 * free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * any later version.
 *
 * The GNU General Public License can be found at
 * http://www.gnu.org/copyleft/gpl.html.
 * A copy is found in the textfile
 * webEdition/licenses/webEditionCMS/License.txt
 *
 * @license	http://www.gnu.org/copyleft/gpl.html  GPL
 */

/**
 * Class we_image_resize
 *
 * Provides functions for creating webEdition buttons.
 */
abstract class we_base_imageEdit{
	const IMAGE_EXTENSIONS = 'svgz';
	const MAXSIZE_ON_FITISIDE = false;
	public const GDIMAGE_TYPE = ['.gif' => 'gif', '.jpg' => 'jpg', '.jpeg' => 'jpg', '.png' => 'png', '.jfif' => 'jpg'];

	/*	 * ***********************************************************************
	 * FUNCTIONS
	 * *********************************************************************** */

	/**
	 * Returns values of the function phpinfo().
	 *
	 * @see        gd_info()
	 *
	 * @return     string[]
	 */
	private static function php_info(): array{
		static $phpinfo = [];

		// Check if need to get the requested information
		if(empty($phpinfo)){
			// Start output buffering
			ob_start();

			phpinfo();

			// Read output of the function phpinfo()
			$returned_phpinfo = ob_get_clean() ?: '';

			// Fill informations of PHP
			$phpinfo = explode("\n", $returned_phpinfo);
		}

		return $phpinfo;
	}

	/**
	 *
	 * @return string[]
	 */
	public static function supported_image_types(): array{
		$output_formats = [];
		$imagetypes = imagetypes();

		// Check output formats
		if($imagetypes & IMG_GIF){
			$output_formats[] = 'gif';
		}

		if($imagetypes & IMG_PNG){
			$output_formats[] = 'png';
		}

		if($imagetypes & IMG_JPG){
			$output_formats[] = 'jpg';
		}

		if($imagetypes & IMG_WEBP){
			$output_formats[] = 'webp';
		}

		return $output_formats;
	}

	public static function detect_image_type(string $filename, string $imagedata = ''): string{
		// Check if we need to read the beginning of the image
		$imagedata = (file_exists($filename) ? we_base_file::loadPart($filename, 0, 20) : substr($imagedata, 0, 20));

		switch(substr($imagedata, 0, 3)){
			case 'GIF':
				return 'gif';
			case "\xFF\xD8\xFF":
				return 'jpg';
			case "\x89" . 'PN':
				return 'png';
			case 'RIF':
				return (substr($imagedata, 8, 4) === 'WEBP') ? 'webp' : '';
			default:
				if(substr($imagedata, 0, 2) === 'BM'){
					return 'bmp';
				}
				return '';
		}
	}

	private static function gd_info(): array{
		// Check if we need to emulate this function since it is built into PHP v4.3.0+ (with bundled GD2 library)
		if(function_exists('gd_info')){
			return gd_info();
		}
		static $gdinfo = [
			'GD Version' => '',
			'FreeType Support' => false,
			'FreeType Linkage' => '',
			'T1Lib Support' => false,
			'GIF Read Support' => false,
			'GIF Create Support' => false,
			'JPG Support' => false,
			'PNG Support' => false,
			'WBMP Support' => false,
			'WebP Support' => false,
			'XBM Support' => false
		];
		if(!empty($gdinfo['GD Version'])){
			return $gdinfo;
		}

		// Now we need to read the phpinfo() to detect the GD library support
		$phpinfo = self::php_info();
		'@phan-var string[] $phpinfo';

		foreach($phpinfo as $value){
			$value = trim(strip_tags($value));

			foreach(array_keys($gdinfo) as $key){
				if(strpos($value, $key) === 0){
					$new_value = trim(str_replace($key, '', $value));
					$gdinfo[$key] = $new_value;
				}
			}
		}

		// Check if GD version information is present now
		if(empty($gdinfo['GD Version'])){
			// Check if we can detect GD library by bypassing the function: phpinfo()

			$imagetypes = imagetypes();

			// Check JPG support
			if($imagetypes & IMG_JPG){
				$gdinfo['JPG Support'] = true;
			}

			// Check PNG support
			if($imagetypes & IMG_PNG){
				$gdinfo['PNG Support'] = true;
			}

			// Check GIF support
			if($imagetypes & IMG_GIF){
				$gdinfo['GIF Create Support'] = true;
			}
			if($imagetypes & IMG_WEBP){
				$gdinfo['WebP Support'] = true;
			}

			// Detect capabilities of GIF support
			if(function_exists('imagecreatefromgif')){
				if(($tempfilename = we_base_file::saveTemp(base64_decode('R0lGODlhAQABAIAAAH//AP///ywAAAAAAQABAAACAUQAOw==')))){//@phan-suppress-current-line PhanPossiblyFalseTypeArgument
					// GIF create support must be enabled if we're able to create a image
					$gif_test = imagecreatefromgif($tempfilename);

					if($gif_test){
						$gdinfo['GIF Read Support'] = true;
					}
					we_base_file::deleteLocalFile($tempfilename);
				}
			}

			// Detect version of GD library
			if(function_exists('imagecreatetruecolor') && @imagecreatetruecolor(1, 1)){
				$gdinfo['GD Version'] = '2.0.1 or higher (assumed)';
			}
		}

		return $gdinfo;
	}

	public static function gd_version(): string{
		static $gdversion = '0';
		// Check if need to get the requested information
		if($gdversion){
			return $gdversion;
		}
		// Request information about GD libary
		$gdinfo = self::gd_info();

		// Define string to be searched
		$searchstring = 'bundled (';

		// Detect information string now
		$gdversion = (substr($gdinfo['GD Version'], 0, strlen($searchstring)) == $searchstring ?
			substr($gdinfo['GD Version'], strlen($searchstring), 3) :
			substr($gdinfo['GD Version'], 0, 3));

		return $gdversion ?: '';
	}

	/**
	 * @param $filename
	 * @return resource
	 */
	public static function ImageCreateFromFileReplacement(string $filename)/* resource */{
		switch(self::detect_image_type($filename)){
			case 'gif':
				if(!function_exists('imagecreatefromgif')){
					throw new we_exception_image('gd not found');
				}
				$ret = imagecreatefromgif($filename);
				break;
			case 'jpg':
				if(!function_exists('imagecreatefromjpeg')){
					throw new we_exception_image('gd not found');
				}
				$ret = imagecreatefromjpeg($filename);
				break;
			case 'png':
				if(!function_exists('imagecreatefrompng')){
					throw new we_exception_image('gd not found');
				}
				$ret = imagecreatefrompng($filename);
				break;
			case 'webp':
				if(!function_exists('imagecreatefromwebp')){
					throw new we_exception_image('gd not found');
				}
				$ret = imagecreatefromwebp($filename);
			default:
				throw new we_exception_image('no function to handle datatype');
		}
		if(!$ret){
			throw new we_exception_image('file not valid');
		}

		return $ret;
	}

	public static function ImageWriteToFileReplacement(/* resource */ $image, string $filename): void{
		switch(self::detect_image_type($filename)){
			case 'gif':
				$success = imagegif($image, $filename);
				break;
			case 'jpg':
				$success = imagejpeg($image, $filename);
				break;
			case 'png':
				$success = imagepng($image, $filename);
				break;
			default:
				throw new we_exception_base('no function to handle datatype');
		}
		if(!$success){
			throw new we_exception_base('resource not valid');
		}
	}

	/**
	 *
	 * @param int $origwidth
	 * @param int $origheight
	 * @param int $newwidth
	 * @param int $newheight
	 * @param bool $keep_aspect_ratio
	 * @param bool $maxsize
	 * @param bool $fitinside
	 * @return array{width:int,height:int,useorig:bool}
	 */
	private static function calculateImageSize(int $origwidth, int $origheight, int $newwidth, int $newheight, bool $keep_aspect_ratio = true, bool $maxsize = true, bool $fitinside = false): array{
		if(self::isNoThumbnailNeeded($origwidth, $origheight, $newwidth, $newheight, $maxsize, $fitinside)){
			return ['width' => $origwidth, 'height' => $origheight, 'useorig' => true];
		}

		// If width has been specified set it and compute new height based on source area aspect ratio
		if($newwidth){
			$outsize = [
				'width' => $newwidth,
				'height' => (int) (round($origheight * $newwidth / $origwidth))
			];
		} else {
			// bugfix #2482: preserve aspect ratio for thumbnails with width=0 and height != 0
			$outsize = [
				'width' => (int) (round(($origwidth / $origheight) * $newheight)),
				'height' => (int) ($newheight ?: round($origheight * $newwidth / $origwidth))
			];
		}

		// If height has been specified set it.
		// If width has already been set and the new image is too tall, compute a new width based
		// on aspect ratio - otherwise, use height and compute new width
		if($newheight){
			if($outsize['height'] > $newheight){
				$outsize['width'] = (int) (round($origwidth * $newheight / $origheight));
				$outsize['height'] = $newheight;
			}
		}

		// Check, if we must discard aspect ratio
		if(!$keep_aspect_ratio && ($newwidth) && ($newheight)){
			$outsize['width'] = $newwidth;
			$outsize['height'] = $newheight;
		}

		// Check, if it is supposed to fit inside
		if($fitinside && ($newwidth) && ($newheight)){
			$outsize['width'] = $newwidth;
			$outsize['height'] = $newheight;
		}

		return ['width' => $outsize['width'], 'height' => $outsize['height'], 'useorig' => false];
	}

	/**
	 *
	 * @param int $origwidth
	 * @param int $origheight
	 * @param int $newwidth
	 * @param int $newheight
	 * @param bool $maxsize
	 * @return array{width:int,height:int,useorig:bool}
	 */
	private static function calculateImageSizeFit(int $origwidth, int $origheight, int $newwidth, int $newheight, bool $maxsize = false): array{
		if(!$newwidth || !$newheight){
			t_e('missing target width or length to fit thumnail inside');
			return ['width' => $origwidth, 'height' => $origheight, 'useorig' => true];
		}

		if(self::isNoThumbnailNeeded($origwidth, $origheight, $newwidth, $newheight, $maxsize, true)){
			return ['width' => $origwidth, 'height' => $origheight, 'useorig' => true]; // FIXME: this is not considered anywhere!
		}

		if($maxsize || ($origwidth > $newwidth && $origheight > $newheight)){
			return ['width' => $newwidth, 'height' => $newheight, 'useorig' => false];
		}

		// not maxsize and exactly one site is longer than its target size: do not resize but crop if needed to fit inside
		if($origwidth <= $newwidth){
			return ['width' => $origwidth, 'height' => $newheight, 'useorig' => false];
		}
		if($origheight <= $newheight){
			return ['width' => $newwidth, 'height' => $origheight, 'useorig' => false];
		}

		return ['width' => $origwidth, 'height' => $origheight, 'useorig' => true];
	}

	public static function isNoThumbnailNeeded(int $origwidth, int $origheight, int $newwidth, int $newheight, bool $maxsize = false, bool $fitinside = false): bool{

		if(/* self::MAXSIZE_ON_FITISIDE && */ $fitinside){
			return false;
		}

		return !$maxsize /* && (!$fitinside) */ && ($origwidth <= $newwidth || $newwidth === 0) && ($origheight <= $newheight || $newheight === 0);
	}

	/**
	 * @param string $filename
	 * @return array{int,int,int,string}
	 * @throws we_exception_image
	 */
	public static function getimagesize(string $filename): array{
		$type = self::detect_image_type($filename);

		if(!self::is_imagetype_supported($type)){
			throw new we_exception_image('not supported image type: ' . $type . ' from file ' . $filename);
		}
		$gdimg = self::ImageCreateFromFileReplacement($filename);
		$ct = 0;
		switch($type){
			case 'gif':
				$ct = 1;
				break;
			case 'jpg':
				$ct = 2;
				break;
			case 'png':
				$ct = 3;
				break;
		}
		$w = ImageSX($gdimg);
		$h = ImageSY($gdimg);
		return [$w, $h, $ct, 'width="' . $w . '" height="' . $h . '"'];
	}

	public static function is_imagetype_supported(string $type): bool{
		return in_array($type, self::supported_image_types(), true);
	}

	public static function is_imagetype_read_supported(string $type): bool{
		if(!function_exists('gd_info')){
			return false;
		}

		return in_array($type, self::supported_image_types(), true);
	}

	/**
	 *
	 * @param resource $img
	 * @param float $amount
	 * @param float $radius
	 * @param float $threshold
	 * @return resource
	 */
	private static function UnsharpMask(/* ressouce */$img, float $amount = 80, float $radius = .5, float $threshold = 3)/* resource */{

////////////////////////////////////////////////////////////////////////////////////////////////
////
////                  Unsharp Mask for PHP - version 2.1.1
////
////    Unsharp mask algorithm by Torstein Hønsi 2003-07.
////             thoensi_at_netcom_dot_no.
////               Please leave this notice.
////
///////////////////////////////////////////////////////////////////////////////////////////////
		// $img is an image that is already created within php using
		// imgcreatetruecolor. No url! $img must be a truecolor image.
		// Attempt to calibrate the parameters to Photoshop:
		$amount = min($amount, 500) * 0.016;
		$radius = round(min(abs($radius), 50) * 2);
		$threshold = min($threshold, 255);
		if($radius == 0){
			return $img;
		}
		$w = imagesx($img);
		$h = imagesy($img);
		if(!$w || !$h){
			throw new we_exception_file('Image to edit is broken');
		}
		$imgCanvas = imagecreatetruecolor($w, $h);
		$imgBlur = imagecreatetruecolor($w, $h);
		assert(is_resource($imgBlur) && is_resource($imgCanvas));

		// Gaussian blur matrix:
		//
		//    1    2    1
		//    2    4    2
		//    1    2    1
		//
		//////////////////////////////////////////////////


		$matrix = [
			[1, 2, 1],
			[2, 4, 2],
			[1, 2, 1]
		];
		imagecopy($imgBlur, $img, 0, 0, 0, 0, $w, $h);
		imageconvolution($imgBlur, $matrix, 16, 0);

		if($threshold > 0){
			// Calculate the difference between the blurred pixels and the original
			// and set the pixels
			for($x = 0; $x < $w - 1; $x++){ // each row
				for($y = 0; $y < $h; $y++){ // each pixel
					$rgbOrig = ImageColorAt($img, $x, $y);
					$rOrig = (($rgbOrig >> 16) & 0xFF);
					$gOrig = (($rgbOrig >> 8) & 0xFF);
					$bOrig = ($rgbOrig & 0xFF);

					$rgbBlur = ImageColorAt($imgBlur, $x, $y);

					$rBlur = (($rgbBlur >> 16) & 0xFF);
					$gBlur = (($rgbBlur >> 8) & 0xFF);
					$bBlur = ($rgbBlur & 0xFF);

					// When the masked pixels differ less from the original
					// than the threshold specifies, they are set to their original value.
					$rNew = (abs($rOrig - $rBlur) >= $threshold) ? max(0, min(255, ($amount * ($rOrig - $rBlur)) + $rOrig)) : $rOrig;
					$gNew = (abs($gOrig - $gBlur) >= $threshold) ? max(0, min(255, ($amount * ($gOrig - $gBlur)) + $gOrig)) : $gOrig;
					$bNew = (abs($bOrig - $bBlur) >= $threshold) ? max(0, min(255, ($amount * ($bOrig - $bBlur)) + $bOrig)) : $bOrig;

					if(($rOrig != $rNew) || ($gOrig != $gNew) || ($bOrig != $bNew)){
						$pixCol = (int) ImageColorAllocate($img, (int) $rNew, (int) $gNew, (int) $bNew);
						ImageSetPixel($img, $x, $y, $pixCol);
					}
				}
			}
		} else {
			for($x = 0; $x < $w; $x++){ // each row
				for($y = 0; $y < $h; $y++){ // each pixel
					$rgbOrig = ImageColorAt($img, $x, $y);
					$rOrig = (($rgbOrig >> 16) & 0xFF);
					$gOrig = (($rgbOrig >> 8) & 0xFF);
					$bOrig = ($rgbOrig & 0xFF);

					$rgbBlur = ImageColorAt($imgBlur, $x, $y);

					$rBlur = (($rgbBlur >> 16) & 0xFF);
					$gBlur = (($rgbBlur >> 8) & 0xFF);
					$bBlur = ($rgbBlur & 0xFF);

					$rNew = ($amount * ($rOrig - $rBlur)) + $rOrig;
					if($rNew > 255){
						$rNew = 255;
					} elseif($rNew < 0){
						$rNew = 0;
					}
					$gNew = ($amount * ($gOrig - $gBlur)) + $gOrig;
					if($gNew > 255){
						$gNew = 255;
					} elseif($gNew < 0){
						$gNew = 0;
					}
					$bNew = ($amount * ($bOrig - $bBlur)) + $bOrig;
					if($bNew > 255){
						$bNew = 255;
					} elseif($bNew < 0){
						$bNew = 0;
					}
					$rgbNew = ($rNew << 16) + ($gNew << 8) + $bNew;
					ImageSetPixel($img, $x, $y, (int) $rgbNew);
				}
			}
		}
		imagedestroy($imgCanvas);
		imagedestroy($imgBlur);

		return $img;
	}

	/**
	 *
	 * @param string $imagedata
	 * @param string $output_format
	 * @param string $output_filename
	 * @param int $output_quality
	 * @param int $width
	 * @param int $height
	 * @param string[] $options
	 * @param array{int,int} $crop
	 * @param int $rotate_angle
	 * @return array{string,int,int} returns filename or data and new width,height
	 * @throws we_exception_file
	 * @throws we_exception_image
	 */
	public static function edit_image(string $imagedata, string $output_format = 'jpg', string $output_filename = '', int $output_quality = 75, int $width = 0, int $height = 0, array $options = [
			we_contents_thumbnail::OPTION_RATIO, we_contents_thumbnail::OPTION_INTERLACE], array $crop = [0, 0], int $rotate_angle = 0): array{
		if(!in_array($output_format, self::supported_image_types(), true)){
			throw new we_exception_image('File format not supported: ' . $output_format);
		}

		$imagepath = (($writeTmpfile = !(strlen($imagedata) < 255 && @file_exists($imagedata))) ?
			we_base_file::saveTemp($imagedata) :
			$imagedata);
		try{
			$gdimg = self::ImageCreateFromFileReplacement($imagepath);
			'@phan-var resource $gdimg';
			$imageInfo = [];
			$imagesize = getimagesize($imagepath, $imageInfo); // $imageinfo contains jpg metadata if exist
			$metasPng = self::extractMetadataFromPng($imagepath);
			unset($imagedata);
		} catch (we_exception_image $e){
			throw $e;
		} finally {
			if($writeTmpfile){
				we_base_file::deleteLocalFile($imagepath);
			}
		}
		// Now we need to ensure that we could read the file
		if(!($imagesize[0] && $imagesize[1])){
			throw new we_exception_image('image not readable');
		}
		switch($output_format){
			case 'jpeg':
				$output_format = 'jpg';
			//no break;
			case 'jpg':
				// Keep quality between 1 and 99
				$output_quality = max(1, min(99, $output_quality));
		}

		$options = array_filter($options);

		$gdWidth = (int) $imagesize[0];
		$gdHeight = (int) $imagesize[1];

		$optFitinside = in_array(we_contents_thumbnail::OPTION_FITINSIDE, $options, true);
		$optMaxsize = in_array(we_contents_thumbnail::OPTION_MAXSIZE, $options, true) || ($optFitinside && self::MAXSIZE_ON_FITISIDE);

		if(($rotate_angle != 0) && function_exists('ImageRotate')){
			$rotate_angle = floatval($rotate_angle);
			if($rotate_angle < 0){
				$rotate_angle = $rotate_angle + ((abs(intval($rotate_angle / 360)) + 1) * 360);
			} elseif($rotate_angle > 360){
				$rotate_angle = $rotate_angle % 360;
			}

			if($rotate_angle > 0){
				$gdimg = ImageRotate($gdimg, $rotate_angle, 0);
				if($gdimg === false){
					throw new we_exception_image('Error while creating gd image');
				}
				$gdWidth = ImageSX($gdimg);
				$gdHeight = ImageSY($gdimg);
				if(!$gdWidth || !$gdHeight){
					throw new we_exception_image('Error while creating gd image');
				}
			}
		}

		if($optFitinside){
			$outsize = self::calculateImageSizeFit($gdWidth, $gdHeight, $width, $height, $optMaxsize);
			$optFitinside = $outsize['useorig'] ? false : $optFitinside;
		} else {
			$outsize = self::calculateImageSize($gdWidth, $gdHeight, $width, $height, in_array(we_contents_thumbnail::OPTION_RATIO, $options, true), true, in_array(we_contents_thumbnail::OPTION_FITINSIDE, $options, true));
		}

		$outsize['width'] = (int) max(1, $outsize['width']);
		$outsize['height'] = (int) max(1, $outsize['height']);

		// Now create the image
		$output_gdimg = (self::gd_version() >= 2.0 ? // this image is always black
			imagecreatetruecolor($outsize['width'], $outsize['height']) :
			imagecreate($outsize['width'], $outsize['height'])
			);
		assert(is_resource($output_gdimg));

		// preserve transparency of png and gif images:
		switch($output_format){
			case 'gif':
				$colorTransparent = imagecolortransparent($gdimg) ?: 0;
				imagepalettecopy($gdimg, $output_gdimg);
				imagefill($output_gdimg, 0, 0, $colorTransparent);
				imagecolortransparent($output_gdimg, $colorTransparent);
				imagetruecolortopalette($output_gdimg, true, 256);
				break;
			case 'png':
				imagealphablending($output_gdimg, false);
				//$transparent = imagecolorallocatealpha($output_gdimg, 0, 0, 0, 127);
				$transparent = imagecolorallocatealpha($output_gdimg, 255, 255, 255, 127) ?: 0;
				imagefill($output_gdimg, 0, 0, $transparent);
				imagesavealpha($output_gdimg, true);
				break;
			default:
		}

		if(($optFitinside || in_array(we_contents_thumbnail::OPTION_CROP, $options, true)) && $width && $height){
			if($optFitinside){
				$width = $outsize['width'];
				$height = $outsize['height'];

				$wratio = $width / $gdWidth;
				$hratio = $height / $gdHeight;
				$ratio = max($width / $gdWidth, $height / $gdHeight);
				$h = (int) ($height / $ratio);
				$w = (int) ($width / $ratio);

				if($wratio < $hratio){
					$x = (int) (($gdWidth - $w) / 2);
					$y = 0;
				} else {
					$x = 0;
					$y = (int) (($gdHeight - $h) / 2);
				}
			} else {
				$h = $height;
				$w = $width;
				$x = (int) (($gdWidth - $w) / 2);
				$y = (int) (($gdHeight - $h) / 2);
			}

			if(array_filter($crop)){
				$xNew = $x + ($w / 2); // x + origthumbwidth/2 => x-Mittelpunkt
				$x = (int) ($xNew + ($xNew * $crop[0]) - ($w / 2)); // Mittelpunkt + Bildfokus - origthumbwidth/2 => Neuer x-Punkt
				if($x + $w > $gdWidth){
					$x = (int) ($gdWidth - $w);
				}
				$x = max(0, $x);
				$yNew = $y + ($h / 2); // y + origthumbheight/2 => y-Mittelpunkt
				$y = (int) ($yNew + ($yNew * $crop[1]) - ($h / 2)); // Mittelpunkt + Bildfokus - origthumbheight/2 => Neuer y-Punkt
				if($y + $h > $gdHeight){
					$y = (int) ($gdHeight - $h);
				}
				$y = max(0, $y);
			}

			if(function_exists('imagecopyresampled')){
				imagecopyresampled($output_gdimg, $gdimg, 0, 0, $x, $y, $width, $height, $w, $h);
			} else {
				imagecopyresized($output_gdimg, $gdimg, 0, 0, $x, $y, $width, $height, $w, $h);
			}
		} elseif(function_exists('imagecopyresampled')){
			imagecopyresampled($output_gdimg, $gdimg, 0, 0, 0, 0, $outsize['width'], $outsize['height'], $gdWidth, $gdHeight);
		} else {
			imagecopyresized($output_gdimg, $gdimg, 0, 0, 0, 0, $outsize['width'], $outsize['height'], $gdWidth, $gdHeight);
		}

		if(in_array(we_contents_thumbnail::OPTION_GAUSSBLUR, $options, true)){
			imagefilter($output_gdimg, IMG_FILTER_GAUSSIAN_BLUR);
		}
		if(in_array(we_contents_thumbnail::OPTION_GRAY, $options, true) || in_array(we_contents_thumbnail::OPTION_SEPIA, $options, true)){
			imagefilter($output_gdimg, IMG_FILTER_GRAYSCALE);
		}
		if(in_array(we_contents_thumbnail::OPTION_NEGATE, $options, true)){
			imagefilter($output_gdimg, IMG_FILTER_NEGATE);
		}
		if(in_array(we_contents_thumbnail::OPTION_SEPIA, $options, true)){
			imagefilter($output_gdimg, IMG_FILTER_COLORIZE, 90, 60, 40);
		}


		if(in_array(we_contents_thumbnail::OPTION_UNSHARP, $options, true)){
			$output_gdimg = self::UnsharpMask($output_gdimg);
		}
		ImageInterlace($output_gdimg, (in_array(we_contents_thumbnail::OPTION_INTERLACE, $options, true) ? 1 : 0));
		$ret = '';
		switch($output_format){
			case 'jpg':
				ob_start();
				imagejpeg($output_gdimg, null, $output_quality);
				try{
					$imageData = self::insertExifIptcToJpeg(ob_get_clean() ?: '', $imageInfo);
					// Output to a filename or directly
					if($output_filename){
						$ret = we_base_file::save($output_filename, $imageData) ? $output_filename : '';
					} elseif(($tempfilename = tempnam(TEMP_PATH, ''))){
						we_base_file::insertIntoCleanUp($tempfilename);
						$ret = $imageData;

						// As we read the temporary file we no longer need it
						we_base_file::deleteLocalFile($tempfilename);
					}
				} catch (we_exception_file $e){
					t_e(sprintf(g_l('weClass', '[edit_image_errorInsertMeta]'), $e->getMessage()));
				}

				break;
			case 'png':
			case 'gif':
				// Set output function
				$image_out_function = 'image' . $output_format;
				// Output to a filename or directly
				if($output_filename){
					if($image_out_function($output_gdimg, $output_filename) && $output_format === 'png' && $metasPng){
						$ret = self::insertMetadataToPng($output_filename, $metasPng) ? $output_filename : '';
					}
				} elseif(($tempfilename = tempnam(TEMP_PATH, ''))){
					$image_out_function($output_gdimg, $tempfilename);
					we_base_file::insertIntoCleanUp($tempfilename, 3600);
					if($output_format === 'png' && $metasPng){
						self::insertMetadataToPng($tempfilename, $metasPng);
					}
					$ret = we_base_file::load($tempfilename); //throws exception
					we_base_file::deleteLocalFile($tempfilename);
				}

				break;
		}

		ImageDestroy($output_gdimg);

		return [$ret, $outsize['width'], $outsize['height']];
	}

	public static function createPreviewThumb(string $imgSrc, int $imgID, int $width, int $height, string &$outputFormat = 'jpg', int $outputQuality = 75, string $tmpName = ''): string{
		if(self::gd_version() == 0){
			$outputFormat = 'gif';
			return WE_DOCUMENT_ROOT . ICON_DIR . 'no_image.gif';
		}
		if(substr($imgSrc, 0, strlen(WE_DOCUMENT_ROOT)) == WE_DOCUMENT_ROOT){ // it is no src, it is a server path
			$imgSrc = substr($imgSrc, strlen(WE_DOCUMENT_ROOT)) ?: '';
		}
		$imgSrc = '/' . ltrim($imgSrc, '/');

		$imgPath = WEBEDITION_PATH . '..' . $imgSrc;
		$path_parts = pathinfo($imgPath);
		if(isset($path_parts['extension']) && ( $path_parts ['extension'] === 'svg' || $path_parts['extension'] === 'svgz')){
			if(file_exists($imgPath)){
				$outputFormat = 'svg-xml';
				return $imgPath;
			}
			$outputFormat = 'gif';
			return WE_DOCUMENT_ROOT . ICON_DIR . 'no_image.gif';
		}
		if(!file_exists($imgPath) || !($imagesize = getimagesize($imgPath))){
			return $imgPath;
		}

		if($imagesize[0] > $width || $imagesize[1] > $height){
			$thumbSrc = ($imgID ?
				WE_PREVIEW_DIR . $imgID . '_' . $width . '_' . $height . '.' . strtolower($outputFormat) :
				TEMP_DIR . ($tmpName ?: we_base_file::getUniqueId()) . '.' . strtolower($outputFormat));
			$thumbPath = WEBEDITION_PATH . '../' . $thumbSrc;

			$thumbExists = file_exists($thumbPath);

			$imageCreationDate = filemtime($imgPath);
			$thumbCreationDate = $thumbExists ? filemtime($thumbPath) : 0;

			if(!$thumbExists || ($imageCreationDate > $thumbCreationDate)){
				self::edit_image($imgPath, $outputFormat, $thumbPath, $outputQuality, $width, $height);
			}
			$thumbSrc = WE_DOCUMENT_ROOT . $thumbSrc;
			//keep the file for 1h
			we_base_file::insertIntoCleanUp($thumbSrc, ($imgID ? 60 * 24 * 3600 : 3600));
			return $thumbSrc;
		}

		return $imgPath;
	}

	/**
	 * returns the HTML for a quality output select box
	 *
	 * @return string
	 * @param string $name
	 * @param int $sel
	 */
	public static function qualitySelect(string $name = 'quality', int $sel = 8): string{
		return '<select name="' . $name . '" class="weSelect">
<option value="0"' . (($sel === 0) ? ' selected="selected"' : '') . '>0 - ' . g_l('weClass', '[quality_low]') . '</option>
<option value="1"' . (($sel === 1) ? ' selected="selected"' : '') . '>1</option>
<option value="2"' . (($sel === 2) ? ' selected="selected"' : '') . '>2</option>
<option value="3"' . (($sel === 3) ? ' selected="selected"' : '') . '>3</option>
<option value="4"' . (($sel === 4) ? ' selected="selected"' : '' ) . '>4 - ' . g_l('weClass', '[quality_medium]') . '</option>
<option value="5"' . (($sel === 5) ? ' selected="selected"' : '') . '>5</option>
<option value="6"' . (($sel === 6) ? ' selected="selected"' : '') . '>6</option>
<option value="7"' . (($sel === 7) ? ' selected="selected"' : '') . '>7</option>
<option value="8"' . (($sel === 8) ? ' selected="selected"' : '') . '>8 - ' . g_l('weClass', '[quality_high]') . '</option>
<option value="9"' . (($sel === 9) ? ' selected="selected"' : '') . '>9</option>
<option value="10"' . (($sel === 10) ? ' selected="selected"' : '' ) . '>10 - ' . g_l('weClass', '[quality_maximum]') . '</option>
</select>';
	}

	private static function getOrdStr(string $string): int{
		$ret = 0;
		for($i = 0; $i < strlen($string); $i++){
			$ret = $ret << 8 | ord($string[$i]);
		}
		return $ret;
	}

	private static function insertExifIptcToJpeg(string $imageData, array $imageInfo = []): string{
		if(!$imageData || !$imageInfo || (empty($imageInfo['APP1']) && empty($imageInfo['APP13']))){
			return $imageData;
		}
		$exif = !empty($imageInfo['APP1']) ? $imageInfo['APP1'] : '';
		if($exif){
			$exifLength = strlen($exif) + 2;
			if($exifLength > 0xFFFF){
				throw new we_exception_file('incompatible exif data found');
			}
			$exif = chr(0xFF) . chr(0xE1) . chr(($exifLength >> 8) & 0xFF) . chr($exifLength & 0xFF) . $exif;
		}

		$iptc = !empty($imageInfo['APP13']) ? $imageInfo['APP13'] : '';
		if($iptc){
			$iptcLength = strlen($iptc) + 2;
			if($iptcLength > 65535){
				throw new we_exception_file('incompatible iptc data found');
			}
			$iptc = chr(0xFF) . chr(0xED) . chr(($iptcLength >> 8) & 0xFF) . chr($iptcLength & 0xFF) . $iptc;
		}

		$imageData = (string) substr($imageData, 2);
		$newHead = chr(0xFF) . chr(0xD8);

		$exifAdded = $exif === '';
		$iptcAdded = $iptc === '';

		while((self::getOrdStr((string) substr($imageData, 0, 2)) & 0xFFF0) === 0xFFE0){ // we are in an APPn segment [E0, EF]
			$segmentLength = (self::getOrdStr((string) substr($imageData, 2, 2)) & 0xFFFF);
			$appNumber = (self::getOrdStr((string) substr($imageData, 1, 1)) & 0x0F);

			if($segmentLength <= 2){ // a valid segment contains at least 2 bytes
				throw new we_exception_file('More than 2 segments found. This is invalid!');
			}

			if(($appNumber >= 1) && (!$exifAdded)){
				$newHead .= $exif;
				$exifAdded = true;
			}

			if(($appNumber >= 13) && (!$iptcAdded)){
				$newHead .= $iptc;
				$iptcAdded = true;
			}

			$newHead .= in_array($appNumber, [1, 13]) ? '' : substr($imageData, 0, $segmentLength + 2); // we replace all APP1 and APP13, even if there are no replacements
			$imageData = substr($imageData, $segmentLength + 2);
		}

		$newHead .= !$exifAdded ? $exif : ''; // add exif if no APP1 was found
		$newHead .= !$iptcAdded ? $iptc : ''; // add iptc if no APP13 was found

		return $newHead . $imageData;
	}

	/* use $returnReadableData = true to extract Metedata from we_imageDocument */

	public static function extractMetadataFromPng(string $imagepath/* , bool $returnReadableData = false */): string{
		if(!@file_exists($imagepath) || !($fh = fopen($imagepath, 'r'))){
			return '';
		}
		$fileheader = fread($fh, 8);
		if($fileheader !== "\x89PNG\x0d\x0a\x1a\x0a"){
			return '';
		}

		//$readableChunks = [];
		$reinsertableChunks = '';
		while(!feof($fh)){
			$binHead = fread($fh, 8); // Length and Type: 8 bytes
			if($binHead === false || strlen($binHead) !== 8){
				break;
			}
			$head = unpack('Nlength/a4type', $binHead);
			if($head['type'] == 'IEND'){
				break;
			}

			if(in_array($head['type'], ['iTXt', 'tEXt', 'zTXt'], true)){
				$reinsertableChunks .= $binHead;

				$content = fread($fh, (int) $head['length']); // Data: length bytes
				if($content === false || strlen($content) !== $head['length']){
					break;
				}
				$reinsertableChunks .= $content;

				/* if($head['type'] === 'tEXt'){
				  list($key, $val) = explode("\0", $content);
				  $readableChunks[$key] = $val;
				  } */

				$reinsertableChunks .= fread($fh, 4); // CRC, 4 bytes
			} elseif(isset($head['length'])){
				fseek($fh, (int) $head['length'] + 4, SEEK_CUR);
			}
		}
		//t_e($readableChunks);

		return /* $returnReadableData ? $readableChunks : */ $reinsertableChunks;
	}

	/**
	 * @param $file
	 * @param $metas
	 * @return bool|string
	 */
	private static function insertMetadataToPng(string $file, string $metas): bool{
		if(!strlen($metas) || !file_exists($file) || !($fh = fopen($file, 'r'))){
			return false;
		}

		$newBindata = fread($fh, 8); // copy png signagture
		$ihdrHead = fread($fh, 8); // copy IHDR header
		if($newBindata === false || $ihdrHead === false){
			return false;
		}
		$newBindata .= $ihdrHead;
		$head = unpack('Nlength/a4type', $ihdrHead);
		$data = fread($fh, (int) $head['length'] + 4);
		if($data === false){
			return false;
		}
		$newBindata .= $data . // copy IHDR content + CRC
			$metas . // insert metas
			fread($fh, filesize($file) ?: 0); // copy rest of image
		fclose($fh);

		return we_base_file::save($file, $newBindata);
	}

}
