<?php
declare(strict_types=1);

/**
 * webEdition CMS
 *
 * $Rev: 4f156532507d $
 * $Author: Marc Kraemer <marc.kraemer@aldente-it.de> $
 * $Date: 2020-05-06 23:12 +0200 $
 *
 * This source is part of webEdition CMS. webEdition CMS is
 * free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * any later version.
 *
 * The GNU General Public License can be found at
 * http://www.gnu.org/copyleft/gpl.html.
 * A copy is found in the textfile
 * webEdition/licenses/webEditionCMS/License.txt
 *
 * @license	http://www.gnu.org/copyleft/gpl.html  GPL
 */
class we_glossary_tree extends we_tree_base{
	public const Types = [
		we_glossary_glossary::TYPE_ABBREVATION,
		we_glossary_glossary::TYPE_ACRONYM,
		we_glossary_glossary::TYPE_FOREIGNWORD,
		we_glossary_glossary::TYPE_LINK,
		we_glossary_glossary::TYPE_TEXTREPLACE,
	];

	protected function customJSFile(): string{
		return we_html_element::jsScript(WE_JS_MODULES_DIR . 'glossary/glossary_tree.js', 'initTree();');
	}

	public static function getItems(int $ParentId, int $Offset = 0, int $Segment = 500, bool $sort = false): array{
		if($ParentId === 0){
			return self::getLanguages();
		}
		list('lang' => $Language, 'type' => $Type) = self::getLangType($ParentId);
		if(!$Type){
			return self::getTypes($ParentId);
		}
		return self::getItemsFromDB($ParentId, $Language, $Type, $Offset, $Segment);
	}

	/**
	 *
	 * @param int $ParentID
	 * @return array{lang:string,type:string}
	 */
	public static function getLangType(int $ParentID): array{
		$langID = (int) (($ParentID + 1) / (count(self::Types) + 10));
		$Language = we_base_WE::$Langs['FrontendLanguages'][$langID];
		$rest = (($ParentID + 1) % (count(self::Types) + 10));
		if($rest === 0){
			return ['lang' => $Language, 'type' => ''];
		}
		$Type = self::Types[$rest - 1];
		return ['lang' => $Language, 'type' => $Type];
	}

	private static function getLanguages(): array{
		$Items = [];
		$pos = -1;
		$decr = (count(self::Types) + 10);
		foreach(we_base_country::getWeFrontendLanguagesForBackend() as $Val){
			$Items[] = [
				'id' => $pos,
				'parentid' => 0,
				'text' => $Val,
				'typ' => 'group',
				'open' => false,
				'disabled' => false,
				'tooltip' => $Val,
				'offset' => 0,
				'published' => true,
				'cmd' => 'we_glossary_view_folder',
				'contenttype' => we_base_ContentTypes::FOLDER_COMMON
			];
			$pos -= $decr;
		}

		return $Items;
	}

	private static function getTypes(int $ParentID): array{
		$Items = [];
		$pos = $ParentID - 1;
		foreach(self::Types as $type){
			$text = g_l('modules_glossary', '[' . $type . ']');
			$Items[] = [
				'id' => $pos--,
				'parentid' => $ParentID,
				'text' => $text,
				'typ' => 'group',
				'open' => false,
				'disabled' => false,
				'tooltip' => $text,
				'offset' => 0,
				'published' => true,
				'cmd' => 'we_glossary_view_type',
				'contenttype' => we_base_ContentTypes::FOLDER_COMMON
			];
		}

		if(we_base_permission::hasPerm([we_base_perm::GLOSSARY_EDIT_DICTIONARY])){
			$Items[] = [
				'id' => $pos--,
				'parentid' => $ParentID,
				'text' => g_l('modules_glossary', '[exception]'),
				'typ' => 'item',
				'open' => false,
				'disabled' => false,
				'tooltip' => g_l('modules_glossary', '[exception]'),
				'offset' => 0,
				'published' => true,
				'cmd' => 'we_glossary_view_exception',
				'contenttype' => we_base_ContentTypes::GLOSSAR
			];
		}

		return $Items;
	}

	private static function getItemsFromDB(int $ParentID, string $Language, string $Type, int $Offset = 0, int $Segment = 500): array{
		$db = new DB_WE();
		$Items = [];
		$PrevOffset = max(0, $Offset - $Segment);

		if($Offset && $Segment){
			$Items[] = [
				'id' => 'prev_' . $ParentID,
				'parentid' => $ParentID,
				'text' => 'display (' . $PrevOffset . '-' . $Offset . ')',
				'contenttype' => self::CT_UP,
				'table' => GLOSSARY_TABLE,
				'typ' => self::TYP_DOTS,
				'open' => false,
				'disabled' => false,
				'tooltip' => '',
				'offset' => $PrevOffset,
			];
		}

		switch($Type){
			case we_glossary_glossary::TYPE_ABBREVATION:
				$cmd = 'we_glossary_edit_abbreviation';
				break;
			case we_glossary_glossary::TYPE_ACRONYM:
				$cmd = 'we_glossary_edit_acronym';
				break;
			case we_glossary_glossary::TYPE_FOREIGNWORD:
				$cmd = 'we_glossary_edit_foreignword';
				break;
			case we_glossary_glossary::TYPE_LINK:
				$cmd = 'we_glossary_edit_link';
				break;
			case we_glossary_glossary::TYPE_TEXTREPLACE:
				$cmd = 'we_glossary_edit_textreplacement';
				break;
			default:
				$cmd = '';
				break;
		}
		$where = '	Language="' . $db->escape($Language) . '" AND
	Type="' . $db->escape($Type) . '"
';
		$all = $db->getAllq('SELECT
	ID,Type,Language,Text,IsPublished,IsFolder
FROM ' . GLOSSARY_TABLE . '
WHERE ' . $where . '
ORDER BY
	(Text REGEXP "^[0-9]") DESC,abs(Text),Text' .
			($Segment ? ' LIMIT ' . $Offset . ',' . $Segment : ''));
		foreach($all as $rec){
			$Items[] = [
				'id' => (int) $rec['ID'],
				'parentid' => $ParentID,
				'text' => htmlspecialchars($rec['Text']),
				'typ' => 'item',
				'open' => false,
				'disabled' => false,
				'tooltip' => (int) $rec['ID'],
				'offset' => $Offset,
				'published' => ($rec['IsPublished'] ? true : false),
				'contenttype' => ($rec['IsFolder'] ? we_base_ContentTypes::FOLDER_COMMON : we_base_ContentTypes::GLOSSAR),
				'cmd' => $cmd,
			];
		}

		$Total = fI('SELECT COUNT(1) FROM ' . $db->escape(GLOSSARY_TABLE) . ' WHERE ' . $where);

		$NextOffset = $Offset + $Segment;
		if($Segment && ($Total > $NextOffset)){
			$Items[] = [
				'id' => 'next_' . $ParentID,
				'parentid' => $ParentID,
				'text' => 'display (' . $NextOffset . '-' . ($NextOffset + $Segment) . ')',
				'contenttype' => self::CT_DOWN,
				'table' => GLOSSARY_TABLE,
				'typ' => self::TYP_DOTS,
				'open' => false,
				'disabled' => false,
				'tooltip' => '',
				'offset' => $NextOffset,
			];
		}
		return $Items;
	}

}
