<?php
declare(strict_types=1);
/**
 * webEdition CMS
 *
 * $Rev: 8b32dd0a5329 $
 * $Author: Marc Kraemer <marc.kraemer@aldente-it.de> $
 * $Date: 2019-12-02 19:36 +0100 $
 *
 * This source is part of webEdition CMS. webEdition CMS is
 * free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * any later version.
 *
 * The GNU General Public License can be found at
 * http://www.gnu.org/copyleft/gpl.html.
 * A copy is found in the textfile
 * webEdition/licenses/webEditionCMS/License.txt
 *
 * @license	http://www.gnu.org/copyleft/gpl.html  GPL
 */

/**
 * WorkfFlow Document Step definition
 * This class describe document step in workflow process
 */
class we_workflow_documentStep extends we_workflow_base{
	const STATUS_NEW = 'new';
	const STATUS_APPROVED = 'approved';
	const STATUS_CANCELED = 'canceled';
	const STATUS_AUTOPUBLISHED = 'autopublished';

	/** @var int */
	var $ID = 0;

	/** @var int */
	var $workflowStepID = 0;

	/** @var DateTime|null */
	protected $startDate = null;

	/** @var DateTime|null */
	protected $finishDate = null;

	/** @var int */
	protected $workflowDocID = 0;

	/** @var string */
	var $Status = self::STATUS_NEW;

	/** @var we_workflow_documentTask[] list of document tasks	 */
	var $tasks = [];

	/**
	 * Default Constructor
	 *
	 * Can load or create new Workflow Step Definition depends of parameter
	 */
	public function __construct(int $wfDocumentStep = 0){
		parent::__construct(WORKFLOW_DOC_STEP_TABLE);

		$this->persistent_slots = [
			'ID' => we_base_request::INT,
			'workflowDocID' => we_base_request::INT,
			'workflowStepID' => we_base_request::INT,
			'startDate' => we_base_request::DATE_TIME,
			'finishDate' => we_base_request::DATE_TIME,
			'Status' => we_base_request::STRING,
		];

		if($wfDocumentStep){
			$this->load($wfDocumentStep);
		}
	}

	/**
	 * Load data from database
	 *
	 */
	function load(int $id = 0, bool $loadData = true): bool{
		if($id){
			parent::load($id);
			## get tasks for workflow
			$this->tasks = we_workflow_documentTask::getAllTasks($this->ID);
			return true;
		}
		return false;
	}

	/**
	 * Start step, activate it
	 *
	 */
	function start(string $desc = ''): bool{
		$workflowDoc = new we_workflow_document($this->workflowDocID);
		$workflowStep = new we_workflow_step($this->workflowStepID);
		$this->startDate = new DateTime();
		$end = (new DateTime())->modify('+ ' . round($workflowStep->Worktime * 3600) . ' second');

		$msg = we_html_element::b(g_l('modules_workflow', '[todo_next]')) .
			we_html_element::b(self::getCurrentUser('Name')) . ':' .
			we_html_element::p([], strtr($desc, ['<br/>' => "\n"])) .
			we_html_element::p([], we_html_element::b(g_l('modules_workflow', '[' . stripTblPrefix($workflowDoc->document->ContentType === we_base_ContentTypes::OBJECT_FILE ? OBJECT_FILES_TABLE : FILE_TABLE) . '][messagePath]')) . ': ' . g_l('weClass', '[path]') . ': ' . $workflowDoc->document->Path . ' (ID:' . $workflowDoc->document->ID . ')') .
			we_html_element::p([], sprintf(g_l('modules_workflow', '[mail][nextStep][deadline]'), $end->format(g_l('date', '[format][default]'))));
		// set all tasks to pending
		foreach($this->tasks as $cur){
			$workflowTask = new we_workflow_task($cur->workflowTaskID);
			if($workflowTask->userID && $workflowTask->Mail){
				$this->sendMail($workflowTask->userID, g_l('modules_workflow', '[todo_next]') . ($workflowDoc->document->Path ? ': ' . $workflowDoc->document->Path : ''), $msg);
			}
		}
		return true;
	}

	function finish(): void{
		$this->finishDate = new DateTime();
	}

	/**
	 * create all tasks for step
	 */
	function createAllTasks(): void{
		$this->tasks = we_workflow_documentTask::createAllTasks($this->workflowStepID);
	}

	/**
	 * save workflow step in database
	 *
	 */
	function save(bool $force_new = false, bool $rebuild = false, array $skipWrite = []): bool{
		if(!parent::save()){
			return false;
		}

		## save all tasks also ##
		foreach($this->tasks as $v){
			$v->documentStepID = $this->ID;
			$v->save();
		}
		return true;
	}

	function delete(): bool{
		$ret = true;
		foreach($this->tasks as $v){
			$r = $v->delete();
			$ret = $ret && $r;
		}
		$r = parent::delete();
		return $ret && $r;
	}

	function approve(int $uID, string $desc, bool $force = false): bool{
		if($force){
			foreach($this->tasks as &$tv){
				$tv->approve();
			}
			$this->Status = self::STATUS_APPROVED;
			$this->finishDate = new DateTime();
			//insert into document Log
			we_workflow_log::logDocumentEvent($this->workflowDocID, $uID, we_workflow_log::TYPE_APPROVE_FORCE, $desc, $this->ID);
			return true;
		}
		try{
			$i = $this->findTaskByUser($uID);
		} catch (we_exception_base $e){
			return false;
		}
		$this->tasks[$i]->approve();

		$workflowStep = new we_workflow_step($this->workflowStepID);
		if($workflowStep->stepCondition){
			$num = $this->findNumOfFinishedTasks();
			if($num == count($this->tasks)){
				$status = true;
				foreach($this->tasks as $v){
					$status = $status && ($v->Status === we_workflow_documentTask::STATUS_APPROVED);
				}

				if($status){
					$this->Status = self::STATUS_APPROVED;
				}
			}
		} else {
			$this->Status = self::STATUS_APPROVED;
		}
		switch($this->Status){
			case self::STATUS_APPROVED:
			case self::STATUS_CANCELED:
				$this->finishDate = new DateTime();
		}
		//insert into document Log
		we_workflow_log::logDocumentEvent($this->workflowDocID, $uID, we_workflow_log::TYPE_APPROVE, $desc, $this->ID);
		return true;
	}

	public function autopublish(int $uID, string $desc, bool $force = false): bool{
		if($force){
			foreach($this->tasks as &$tv){
				$tv->approve();
			}
			$this->Status = self::STATUS_APPROVED;
			$this->finishDate = new DateTime();
			//insert into document Log
			we_workflow_log::logDocumentEvent($this->workflowDocID, $uID, we_workflow_log::TYPE_APPROVE_FORCE, $desc, $this->ID);
			return true;
		}
		try{
			$i = $this->findTaskByUser($uID);
		} catch (we_exception_base $e){
			return false;
		}

		$this->tasks[$i]->approve();

		$workflowStep = new we_workflow_step($this->workflowStepID);
		if($workflowStep->stepCondition){
			$num = $this->findNumOfFinishedTasks();
			if($num == count($this->tasks)){
				$status = true;
				foreach($this->tasks as $v){
					$status &= ($v->Status === we_workflow_documentTask::STATUS_APPROVED);
				}

				if($status){
					$this->Status = self::STATUS_APPROVED;
				}
			}
		} else {
			$this->Status = self::STATUS_APPROVED;
		}
		switch($this->Status){
			case self::STATUS_APPROVED:
			case self::STATUS_CANCELED:
				$this->finishDate = new DateTime();
		}
		//insert into document Log
		we_workflow_log::logDocumentEvent($this->workflowDocID, $uID, we_workflow_log::TYPE_APPROVE, $desc, $this->ID);
		return true;
	}

	public function decline(int $uID, string $desc, bool $force = false): bool{
		if($force){
			foreach($this->tasks as &$tv){
				$tv->decline();
			}
			$this->Status = self::STATUS_CANCELED;
			$this->finishDate = new DateTime();
			//insert into document Log
			we_workflow_log::logDocumentEvent($this->workflowDocID, $uID, we_workflow_log::TYPE_DECLINE, $desc, $this->ID);
			return true;
		}

		try{
			$i = $this->findTaskByUser($uID);
		} catch (we_exception_base $e){
			return false;
		}

		$this->tasks[$i]->decline();
		//FIXME: since next var is unused, does this operation do anything except ressource usage?
		//$workflowStep = new we_workflow_step($this->workflowStepID);
		$this->Status = self::STATUS_CANCELED;
		switch($this->Status){
			case self::STATUS_APPROVED:
			case self::STATUS_CANCELED:
				$this->finishDate = new DateTime();
		}
		//insert into document Log
		we_workflow_log::logDocumentEvent($this->workflowDocID, $uID, we_workflow_log::TYPE_DECLINE, $desc, $this->ID);
		return true;
	}

	function findTaskByUser(int $uID): int{
		foreach($this->tasks as $i => $task){
			$workflowTask = new we_workflow_task($task->workflowTaskID);
			if($workflowTask->userID == $uID){
				return (int) $i;
			}
		}
		throw new we_exception_base('not found');
	}

	function findNumOfFinishedTasks(): int{
		$num = 0;
		foreach($this->tasks as $task){
			if($task->Status != we_workflow_documentTask::STATUS_NEW){
				$num++;
			}
		}
		return $num;
	}

	/**
	 * return all steps for workflow document (created)
	 * @param int $workflowDocumentID
	 * @return we_workflow_documentStep[]
	 */
	static function getAllSteps(int $workflowDocumentID): array{
		$all = DB_WE::$inst->getAllq('SELECT ID FROM ' . WORKFLOW_DOC_STEP_TABLE . ' WHERE workflowDocID=' . $workflowDocumentID . ' ORDER BY ID', true);
		'@phan-var int[] $all';
		$docSteps = [];
		foreach($all as $cur){
			$docSteps[] = new self($cur);
		}
		return $docSteps;
	}

	/**
	 * create all steps for workflow document
	 *
	 */
	static function createAllSteps(int $workflowID): array{
		$all = DB_WE::$inst->getAllq('SELECT * FROM ' . WORKFLOW_STEP_TABLE . ' WHERE workflowID=' . $workflowID . ' ORDER BY ID');
		'@phan-var array $all';
		$docSteps = [];
		foreach($all as $cur){
			$docSteps[] = self::createStepFromHash($cur);
		}
		return $docSteps;
	}

	/**
	 * Create step from hash
	 *
	 */
	private static function createStepFromHash(array $WorkflowStepArray): self{
		$docStep = new self();

		$docStep->workflowStepID = $WorkflowStepArray['ID'];
		$docStep->tasks = we_workflow_documentTask::createAllTasks($docStep->workflowStepID);
		return $docStep;
	}

	public function hasStarted(): bool{
		return $this->startDate != null;
	}

	public function getTimeStatus(): array{
		$end = $this->startDate ? clone $this->startDate : new DateTime();
		$end->modify("+" . round((new we_workflow_step($this->workflowStepID))->Worktime * 3600) . ' second');
		$remain = clone $end;

		return [
			'elapsed' => $this->startDate ? (new DateTime())->diff($this->startDate) : 0,
			'remain' => $remain->diff(new DateTime()),
			'end' => $end
		];
	}

	public function setDocID(int $docID): void{
		$this->workflowDocID = $docID;
	}

}
